/*
	Copyright (C) 2011  Andrew Cotter

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/**
	\file output.cc
	\brief Example program which takes an input NODE file, and outputs a well-spaced superset
*/




#include "data.hh"
#include "dynamic_wsp.h"


#include <boost/filesystem.hpp>

#include <boost/array.hpp>
#include <boost/shared_array.hpp>
#include <boost/assert.hpp>
#include <boost/static_assert.hpp>


#include <vector>

#include <iostream>

#include <stdexcept>

#include <cstdint>
#include <cstdlib>
#include <cstddef>
#include <cmath>




//============================================================================
//    main function
//============================================================================


/**
	\brief main function

	This program takes three parameters: a filename, in either PLY or NODE
	format; a well-spacedness parameter rho (must be greater than one); and the
	number of threads to use.

	The given file will be made well-spaced via the insertion of Steiner
	points. The resulting point set will be written to stdout in NODE format,
	first the input points, then the Steiner points.
*/
int main( int argc, char* argv[] ) {

	if ( argc != 4 ) {

		std::cerr << argv[ 0 ] << " <filename> <rho> <threads>" << std::endl;
		exit( EXIT_FAILURE );
	}

	boost::filesystem::path const path( argv[ 1 ] );
	double const rho = atof( argv[ 2 ] );
	int const threads = atoi( argv[ 3 ] );
	if ( ( ! boost::filesystem::is_regular( path ) ) || ( rho <= 1 ) || ( threads < 1 ) ) {

		if ( ! boost::filesystem::is_regular( path ) )
			std::cerr << '"' << path << "\" does not exist" << std::endl;
		exit( EXIT_FAILURE );
	}

	try {

		// read the points
		unsigned int dimension = 0;
		std::vector< boost::shared_array< double > > pPoints;
		ReadFile( dimension, pPoints, path );

		// scale and recenter the pPoints to the bounding box
		{	boost::shared_array< double > lower( new double[ dimension ] );
			boost::shared_array< double > upper( new double[ dimension ] );
			std::fill( lower.get(), lower.get() + dimension,  std::numeric_limits< double >::infinity() );
			std::fill( upper.get(), upper.get() + dimension, -std::numeric_limits< double >::infinity() );
			{	std::vector< boost::shared_array< double > >::const_iterator ii    = pPoints.begin();
				std::vector< boost::shared_array< double > >::const_iterator iiEnd = pPoints.end();
				for ( ; ii != iiEnd; ++ii ) {

					for ( unsigned int jj = 0; jj < dimension; ++jj ) {

						lower[ jj ] = std::min( lower[ jj ], ( *ii )[ jj ] );
						upper[ jj ] = std::max( upper[ jj ], ( *ii )[ jj ] );
					}
				}
			}

			boost::shared_array< double > center( new double[ dimension ] );
			double width = 0;
			for ( unsigned int ii = 0; ii < dimension; ++ii ) {

				center[ ii ] = ( lower[ ii ] + upper[ ii ] ) * 0.5;
				width = std::max( width, upper[ ii ] - lower[ ii ] );
			}
			double const scale = 0.95 / width;    // **TODO: parameter

			{	std::vector< boost::shared_array< double > >::iterator ii    = pPoints.begin();
				std::vector< boost::shared_array< double > >::iterator iiEnd = pPoints.end();
				for ( ; ii != iiEnd; ++ii ) {

					for ( unsigned int jj = 0; jj < dimension; ++jj ) {

						// scale and shift to (0,1)
						( *ii )[ jj ] = ( ( *ii )[ jj ] - center[ jj ] ) * scale + 0.5;
						BOOST_ASSERT( ( *ii )[ jj ] > 0 );
						BOOST_ASSERT( ( *ii )[ jj ] < 1 );
					}
				}
			}
		}

		// set beta based on rho and the dimension
		double beta = rho * 1.99;
		switch( dimension ) {
			case 2: beta = rho * std::sqrt( 2.0 ); break;
			case 3: beta = rho * 2.0 / std::sqrt( 3.0 ); break;
			default: break;
		}

		// create well-spaced point set
		DynamicWSP_Handle handle = NULL;
		if ( DynamicWSP_Allocate( &handle, dimension, rho, beta ) )
			throw std::runtime_error( DynamicWSP_Error() );

		// insert all of the points
		{	std::vector< boost::shared_array< double > >::const_iterator ii    = pPoints.begin();
			std::vector< boost::shared_array< double > >::const_iterator iiEnd = pPoints.end();
			for ( ; ii != iiEnd; ++ii ) {

				DynamicWSP_PointHandle pointHandle = NULL;
				if ( DynamicWSP_Insert( handle, &pointHandle, ii->get() ) ) {

					DynamicWSP_Free( handle );
					throw std::runtime_error( DynamicWSP_Error() );
				}

				if ( pointHandle == NULL )
					std::cerr << "Duplicate point ignored" << std::endl;
			}
		}

		// make the point set well-spaced by inserting Steiner points
		if ( DynamicWSP_Refine( handle, threads ) ) {

			DynamicWSP_Free( handle );
			throw std::runtime_error( DynamicWSP_Error() );
		}

		bool complete;

		// get the handles of all "normal" points
		unsigned int points = 0;
		if ( DynamicWSP_GetPointHandles( handle, NULL, &points, NULL ) ) {

			DynamicWSP_Free( handle );
			throw std::runtime_error( DynamicWSP_Error() );
		}
		boost::shared_array< DynamicWSP_PointHandle > pointHandles( new DynamicWSP_PointHandle[ points ] );
		unsigned int usedPoints = points;
		if ( DynamicWSP_GetPointHandles( handle, pointHandles.get(), &usedPoints, &complete ) ) {

			DynamicWSP_Free( handle );
			throw std::runtime_error( DynamicWSP_Error() );
		}
		BOOST_ASSERT( usedPoints == points );
		BOOST_ASSERT( complete );

		// get the handles of all Steiner points
		unsigned int steinerPoints = 0;
		if ( DynamicWSP_GetSteinerPointHandles( handle, NULL, &steinerPoints, NULL ) ) {

			DynamicWSP_Free( handle );
			throw std::runtime_error( DynamicWSP_Error() );
		}
		boost::shared_array< DynamicWSP_PointHandle > steinerPointHandles( new DynamicWSP_PointHandle[ steinerPoints ] );
		unsigned int usedSteinerPoints = steinerPoints;
		if ( DynamicWSP_GetSteinerPointHandles( handle, steinerPointHandles.get(), &usedSteinerPoints, &complete ) ) {

			DynamicWSP_Free( handle );
			throw std::runtime_error( DynamicWSP_Error() );
		}
		BOOST_ASSERT( usedSteinerPoints == steinerPoints );
		BOOST_ASSERT( complete );

		// output a new NODE file with the normal points first, then the Steiner points
		std::cout << ( points + steinerPoints ) << ' ' << dimension << " 0 0" << std::endl;
		unsigned int index = 1;
		boost::shared_array< double > coordinates( new double[ dimension ] );
		for ( unsigned int ii = 0; ii < points; ++ii ) {

			if ( DynamicWSP_GetCoordinates( handle, pointHandles[ ii ], coordinates.get() ) ) {

				DynamicWSP_Free( handle );
				throw std::runtime_error( DynamicWSP_Error() );
			}

			std::cout << ( index++ );
			for ( unsigned int jj = 0; jj < dimension; ++jj )
				std::cout << ' ' << coordinates[ jj ];
			std::cout << std::endl;
		}
		for ( unsigned int ii = 0; ii < steinerPoints; ++ii ) {

			if ( DynamicWSP_GetCoordinates( handle, steinerPointHandles[ ii ], coordinates.get() ) ) {

				DynamicWSP_Free( handle );
				throw std::runtime_error( DynamicWSP_Error() );
			}

			std::cout << ( index++ );
			for ( unsigned int jj = 0; jj < dimension; ++jj )
				std::cout << ' ' << coordinates[ jj ];
			std::cout << std::endl;
		}

		// free the well-spaced point set
		DynamicWSP_Free( handle );
	}
	catch( std::exception& error ) {

		std::cerr << error.what() << std::endl;
		exit( EXIT_FAILURE );
	}

	return EXIT_SUCCESS;
}
