/*
	Copyright (C) 2011  Andrew Cotter

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/**
	\internal
	\file accuracy_approximate.hh
	\brief Accuracy::Approximate class, and associated functions
*/




#ifndef __ACCURACY_APPROXIMATE_HH__
#define __ACCURACY_APPROXIMATE_HH__

#ifdef __cplusplus




#include "helpers.hh"

#include <boost/assert.hpp>
#include <boost/static_assert.hpp>

#include <ostream>
#include <algorithm>
#include <limits>
#include <cmath>




namespace Accuracy {




namespace {




//============================================================================
//    FloatTraits helper template
//============================================================================


template< typename t_Type >
struct FloatTraits;


template<>
struct FloatTraits< double > {

	BOOST_STATIC_ASSERT( std::numeric_limits< double >::radix == 2 );

	enum {
		SIGN = 1,
		EXPONENT = 11,
		EXPONENT_BIAS = 1023,
		MANTISSA = 52
	};
	BOOST_STATIC_ASSERT( MANTISSA + 1 == std::numeric_limits< double >::digits );
	BOOST_STATIC_ASSERT( SIGN + EXPONENT + MANTISSA == sizeof( double ) * 8 );

	typedef double Type;


	static inline int const Exponent( double const value ) {    // seems to be slightly faster than a floating-point multiply

		BOOST_STATIC_ASSERT( sizeof( uint64_t ) == sizeof( double ) );

		union { uint64_t ii; double ff; } both;
		both.ff = value;
		both.ii &= ( ( 1ull << ( EXPONENT + MANTISSA ) ) - 1 );    // clear the sign bit
		both.ii >>= MANTISSA;
		BOOST_ASSERT( ( value != 0 ) || ( both.ii == 0 ) );

		return( static_cast< int >( both.ii ) - EXPONENT_BIAS );
	}


	// returns a number at *least* as large as 2^power
	static inline double const Construct( int const power ) {

		union { uint64_t ii; double ff; } both;

		if ( power <= -static_cast< int >( EXPONENT_BIAS ) )
			both.ff = std::numeric_limits< double >::min();
		else if ( power >= static_cast< int >( ( 1ull << EXPONENT ) - EXPONENT_BIAS ) )
			both.ff = std::numeric_limits< double >::infinity();
		else
			both.ii = static_cast< uint64_t >( power + EXPONENT_BIAS ) << MANTISSA;

		return both.ff;
	}


	// moves value away from zero by at least 2^power
	static inline double const Increase( double const value, int const power ) {

		BOOST_ASSERT( value != 0 );

		union { uint64_t ii; double ff; } both;
		both.ff = value;

		int     exponent = ( ( both.ii >> MANTISSA ) & ( ( 1ull << EXPONENT ) - 1 ) );
		int64_t mantissa = ( both.ii & ( ( 1ull << MANTISSA ) - 1 ) );
		exponent -= EXPONENT_BIAS + MANTISSA;
		mantissa |= ( 1ull << MANTISSA );

		// mantissa += 1 << ( power - exponent )
		if ( power < exponent )
			++mantissa;
		else if ( power > exponent + MANTISSA ) {

			int const shift = power - ( exponent + MANTISSA );
			if ( shift > MANTISSA )
				mantissa = 0;
			else
				mantissa >>= shift;
			mantissa += ( 1ull << MANTISSA ) + 1;
			exponent = power - MANTISSA;
		}
		else
			mantissa += ( 1ull << ( power - exponent ) );

		// make sure the high bit of mantissa is MANTISSA
		BOOST_ASSERT( mantissa > 0 );
		int highBit = HighBit( static_cast< uint64_t >( mantissa ) );
		BOOST_ASSERT( highBit >= MANTISSA );
		if ( highBit > MANTISSA ) {

			do {

				mantissa >>= ( highBit - MANTISSA );
				++mantissa;
				exponent += ( highBit - MANTISSA );
				highBit = HighBit( static_cast< uint64_t >( mantissa ) );
				BOOST_ASSERT( highBit >= MANTISSA );

			} while ( highBit > MANTISSA );
		}

		// at this point, the only possible problem is that exponent might be outside the valid range
		exponent += EXPONENT_BIAS + MANTISSA;
		mantissa &= ( ( 1ull << MANTISSA ) - 1 );
		BOOST_ASSERT( exponent > 0 );
		if ( exponent >= static_cast< int >( 1ull << EXPONENT ) )
			both.ff = ( both.ff < 0 ) ? -std::numeric_limits< double >::infinity() : std::numeric_limits< double >::infinity();
		else {

			both.ii &= ( 1ull << ( EXPONENT + MANTISSA ) );    // keep only the sign bit
			both.ii |= ( static_cast< uint64_t >( exponent ) << MANTISSA );
			both.ii |= static_cast< uint64_t >( mantissa );
		}

		return both.ff;
	}


	// moves value towards zero by at least 2^power
	static inline double const Decrease( double const value, int const power ) {

		BOOST_ASSERT( value != 0 );

		union { uint64_t ii; double ff; } both;
		both.ff = value;

		int     exponent = ( ( both.ii >> MANTISSA ) & ( ( 1ull << EXPONENT ) - 1 ) );
		int64_t mantissa = ( both.ii & ( ( 1ull << MANTISSA ) - 1 ) );
		exponent -= EXPONENT_BIAS + MANTISSA;
		mantissa |= ( 1ull << MANTISSA );

		// mantissa += 1 << ( power - exponent )
		if ( power < exponent )
			--mantissa;
		else if ( power > exponent + MANTISSA ) {

			int const shift = power - ( exponent + MANTISSA );
			if ( shift > MANTISSA )
				mantissa = 0;
			else
				mantissa >>= shift;
			mantissa = ( 1ull << MANTISSA ) - mantissa;
			exponent = power - MANTISSA;
			both.ii ^= ( 1ull << ( EXPONENT + MANTISSA ) );    // flip the sign
		}
		else
			mantissa -= ( 1ull << ( power - exponent ) );

		BOOST_ASSERT( mantissa >= 0 );
		if ( mantissa == 0 )
			both.ff = 0;
		else {

			// make sure the high bit of mantissa is MANTISSA
			int highBit = HighBit( static_cast< uint64_t >( mantissa ) );
			BOOST_ASSERT( highBit <= MANTISSA );
			if ( highBit < MANTISSA ) {

				mantissa <<= ( MANTISSA - highBit );
				exponent -= ( MANTISSA - highBit );
			}

			// at this point, the only possible problem is that exponent might be outside the valid range
			exponent += EXPONENT_BIAS + MANTISSA;
			mantissa &= ( ( 1ull << MANTISSA ) - 1 );
			if ( exponent <= 0 )
				both.ff = ( both.ff < 0 ) ? -std::numeric_limits< double >::min() : std::numeric_limits< double >::min();
			else if ( exponent >= static_cast< int >( 1ull << EXPONENT ) )
				both.ff = ( both.ff < 0 ) ? -std::numeric_limits< double >::infinity() : std::numeric_limits< double >::infinity();
			else {

				both.ii &= ( 1ull << ( EXPONENT + MANTISSA ) );    // keep only the sign bit
				both.ii |= ( static_cast< uint64_t >( exponent ) << MANTISSA );
				both.ii |= static_cast< uint64_t >( mantissa );
			}
		}

		return both.ff;
	}
};




}    // anonymous namespace




//============================================================================
//    Approximate class
//============================================================================


struct Approximate {

	enum {
		MAXIMUM_ACCURACY = static_cast< int >( 1ull << FloatTraits< double >::EXPONENT ),
		MINIMUM_ACCURACY = 1 - FloatTraits< double >::EXPONENT_BIAS    // smallest possible nonzero value is 2^(1-EXPONENT_BIAS)
	};


	inline Approximate();

	inline explicit Approximate( double const value );

	inline Approximate( double const value, int const log2UpperBound, int const log2Epsilon );

	inline Approximate( Approximate const& other );


	//inline operator double const();


	inline double const GetValue() const;
	inline int const GetLog2UpperBound() const;
	inline int const GetLog2Epsilon() const;

	inline double const GetLowerBound() const;
	inline double const GetUpperBound() const;


	inline Approximate const& operator=( Approximate const& other );


	inline int const Sign() const;    // +1 => positive, -1 => negative, 0 => not sure

	inline void Negate();
	inline void Square();


	inline Approximate const operator-() const;

	inline Approximate const& operator+=( Approximate const& other );
	inline Approximate const& operator+=( double const other );

	inline Approximate const& operator-=( Approximate const& other );
	inline Approximate const& operator-=( double const other );

	inline Approximate const& operator*=( Approximate const& other );
	inline Approximate const& operator*=( double const other );

	inline Approximate const& operator/=( Approximate const& other );
	inline Approximate const& operator/=( double const other );


	static inline int const Compare( Approximate const& left, Approximate const& right );
	static inline int const Compare( Approximate const& left, double const right );
	static inline int const Compare( double const left, Approximate const& right );


private:

	double m_value;
	int m_log2UpperBound;
	int m_log2Epsilon;
};




//============================================================================
//    Approximate inline methods
//============================================================================


Approximate::Approximate() :
	m_value( std::numeric_limits< double >::quiet_NaN() )
{
}


Approximate::Approximate( double const value ) :
	m_value( value ),
	m_log2UpperBound( FloatTraits< double >::Exponent( m_value ) + 1 ),
	m_log2Epsilon( m_log2UpperBound - 1 - FloatTraits< double >::MANTISSA )
{
}


Approximate::Approximate( double const value, int const log2UpperBound, int const log2Epsilon ) :
	m_value( value ),
	m_log2UpperBound( log2UpperBound ),
	m_log2Epsilon( log2Epsilon )
{
}


Approximate::Approximate( Approximate const& other ) :
	m_value( other.m_value ),
	m_log2UpperBound( other.m_log2UpperBound ),
	m_log2Epsilon( other.m_log2Epsilon )
{
}


/*
Approximate::operator double const() {

	return m_value;
}
*/


double const Approximate::GetValue() const {

	return m_value;
}


int const Approximate::GetLog2UpperBound() const {

	return m_log2UpperBound;
}


int const Approximate::GetLog2Epsilon() const {

	return m_log2Epsilon;
}


double const Approximate::GetLowerBound() const {

	double result;
	if ( m_value > 0 )
		result = FloatTraits< double >::Decrease( m_value, m_log2Epsilon );
	else if ( m_value < 0 )
		result = FloatTraits< double >::Increase( m_value, m_log2Epsilon );
	else
		result = -FloatTraits< double >::Construct( m_log2Epsilon );

	return result;
}


double const Approximate::GetUpperBound() const {

	double result;
	if ( m_value > 0 )
		result = FloatTraits< double >::Increase( m_value, m_log2Epsilon );
	else if ( m_value < 0 )
		result = FloatTraits< double >::Decrease( m_value, m_log2Epsilon );
	else
		result = FloatTraits< double >::Construct( m_log2Epsilon );

	return result;
}


Approximate const& Approximate::operator=( Approximate const& other ) {

	m_value = other.m_value;
	m_log2UpperBound = other.m_log2UpperBound;
	m_log2Epsilon = other.m_log2Epsilon;

	return *this;
}


int const Approximate::Sign() const {

	int result = ( m_value < 0 ) ? -1 : 1;
	if (
		UNLIKELY( ! std::isfinite( m_value ) ) ||
		UNLIKELY( m_log2UpperBound <= m_log2Epsilon ) ||
		UNLIKELY( FloatTraits< double >::Exponent( m_value ) <= m_log2Epsilon )
	)
	{
		result = 0;
	}

	return result;
}


void Approximate::Negate() {

	m_value = -m_value;
}


void Approximate::Square() {

	int const term = m_log2Epsilon + m_log2UpperBound;
	int const cross = m_log2Epsilon * 2;

	m_value *= m_value;

	m_log2UpperBound *= 2;
	//m_log2UpperBound = FloatTraits< double >::Exponent( m_value ) + 1;
	BOOST_ASSERT( ( ! std::isfinite( m_value ) ) || ( m_value == 0 ) || ( m_log2UpperBound > FloatTraits< double >::Exponent( m_value ) ) );

	if ( UNLIKELY( cross > term ) )
		m_log2Epsilon = 1 + cross;
	else
		m_log2Epsilon = 2 + term;
}


Approximate const Approximate::operator-() const {

	return Approximate( -m_value, m_log2UpperBound, m_log2Epsilon );
}


Approximate const& Approximate::operator+=( Approximate const& other ) {

	m_value += other.m_value;

	//m_log2UpperBound = 1 + std::max( m_log2UpperBound, other.m_log2UpperBound );
	m_log2UpperBound = FloatTraits< double >::Exponent( m_value ) + 1;
	BOOST_ASSERT( ( ! std::isfinite( m_value ) ) || ( m_value == 0 ) || ( m_log2UpperBound > FloatTraits< double >::Exponent( m_value ) ) );

	m_log2Epsilon = 1 + std::max( m_log2Epsilon, other.m_log2Epsilon );

	return *this;
}


Approximate const& Approximate::operator+=( double const other ) {

	m_value += other;

	m_log2UpperBound = FloatTraits< double >::Exponent( m_value ) + 1;

	m_log2Epsilon = 1 + std::max( m_log2Epsilon, m_log2UpperBound - 1 - FloatTraits< double >::MANTISSA );

	return *this;
}


Approximate const& Approximate::operator-=( Approximate const& other ) {

	m_value -= other.m_value;

	//m_log2UpperBound = 1 + std::max( m_log2UpperBound, other.m_log2UpperBound );
	m_log2UpperBound = FloatTraits< double >::Exponent( m_value ) + 1;
	BOOST_ASSERT( ( ! std::isfinite( m_value ) ) || ( m_value == 0 ) || ( m_log2UpperBound > FloatTraits< double >::Exponent( m_value ) ) );

	m_log2Epsilon = 1 + std::max( m_log2Epsilon, other.m_log2Epsilon );

	return *this;
}


Approximate const& Approximate::operator-=( double const other ) {

	m_value -= other;

	m_log2UpperBound = FloatTraits< double >::Exponent( m_value ) + 1;

	m_log2Epsilon = 1 + std::max(
		m_log2Epsilon,
		m_log2UpperBound - 1 - FloatTraits< double >::MANTISSA
	);

	return *this;
}


Approximate const& Approximate::operator*=( Approximate const& other ) {

	int const term1 = other.m_log2Epsilon +       m_log2UpperBound;
	int const term2 =       m_log2Epsilon + other.m_log2UpperBound;
	int const cross = m_log2Epsilon + other.m_log2Epsilon;

	m_value *= other.m_value;

	m_log2UpperBound += other.m_log2UpperBound;
	BOOST_ASSERT( ( ! std::isfinite( m_value ) ) || ( m_value == 0 ) || ( m_log2UpperBound > FloatTraits< double >::Exponent( m_value ) ) );

	if ( term1 > term2 ) {

		if ( LIKELY( term1 > cross ) )    // term1 > term2, cross
			m_log2Epsilon = 1 + term1;
		else if ( cross > term1 )    // cross > term1, term2
			m_log2Epsilon = 1 + cross;
		else    // term1 == cross > term2
			m_log2Epsilon = 2 + term1;
	}
	else if ( term2 > term1 ) {

		if ( LIKELY( term2 > cross ) )    // term2 > term1, cross
			m_log2Epsilon = 1 + term2;
		else if ( cross > term2 )    // cross > term1, term2
			m_log2Epsilon = 1 + cross;
		else    // term2 == cross > term1
			m_log2Epsilon = 2 + term2;
	}
	else {

		if ( UNLIKELY( cross > term1 ) )    // cross > term1, term2
			m_log2Epsilon = 1 + cross;
		else    // term1 == term2 >= cross
			m_log2Epsilon = 2 + term1;
	}

	return *this;
}


Approximate const& Approximate::operator*=( double const other ) {

	m_value *= other;

	m_log2UpperBound = FloatTraits< double >::Exponent( m_value ) + 1;

	m_log2Epsilon = std::max(
		m_log2Epsilon + ( FloatTraits< double >::Exponent( other ) + 1 ),
		m_log2UpperBound - 1 - FloatTraits< double >::MANTISSA
	);

	return *this;
}


Approximate const& Approximate::operator/=( Approximate const& other ) {

	int const log2LowerBound = FloatTraits< double >::Exponent( other.m_value );
	if ( LIKELY( log2LowerBound > other.m_log2Epsilon ) ) {

		int const term1 = other.m_log2Epsilon +       m_log2UpperBound;
		int const term2 =       m_log2Epsilon + other.m_log2UpperBound;

		m_value /= other.m_value;

		m_log2UpperBound = FloatTraits< double >::Exponent( m_value ) + 1;

		m_log2Epsilon = 2 + std::max( term1, term2 ) - 2 * log2LowerBound;
	}
	else
		m_value = std::numeric_limits< double >::quiet_NaN();

	return *this;
}


Approximate const& Approximate::operator/=( double const other ) {

	m_value /= other;

	m_log2UpperBound = FloatTraits< double >::Exponent( m_value ) + 1;

	m_log2Epsilon = std::max(
		m_log2Epsilon - FloatTraits< double >::Exponent( other ),
		m_log2UpperBound - 1 - FloatTraits< double >::MANTISSA
	);

	return *this;
}


int const Approximate::Compare( Approximate const& left, Approximate const& right ) {

	Approximate result( left );
	result -= right;
	return result.Sign();
}


int const Approximate::Compare( Approximate const& left, double const right ) {

	Approximate result( left );
	result -= right;
	return result.Sign();
}


int const Approximate::Compare( double const left, Approximate const& right ) {

	Approximate result( right );
	result -= left;
	return -result.Sign();
}




}    // namespace Accuracy




//============================================================================
//    Approximate operators
//============================================================================


inline Accuracy::Approximate const operator+( Accuracy::Approximate const& left, Accuracy::Approximate const& right ) {

	Accuracy::Approximate result( left );
	result += right;

	return result;
}


inline Accuracy::Approximate const operator+( Accuracy::Approximate const& left, double const right ) {

	Accuracy::Approximate result( left );
	result += right;

	return result;
}


inline Accuracy::Approximate const operator+( double const left, Accuracy::Approximate const& right ) {

	Accuracy::Approximate result( right );
	result += left;

	return result;
}


inline Accuracy::Approximate const operator-( Accuracy::Approximate const& left, Accuracy::Approximate const& right ) {

	Accuracy::Approximate result( left );
	result -= right;

	return result;
}


inline Accuracy::Approximate const operator-( Accuracy::Approximate const& left, double const right ) {

	Accuracy::Approximate result( left );
	result -= right;

	return result;
}


inline Accuracy::Approximate const operator-( double const left, Accuracy::Approximate const& right ) {

	Accuracy::Approximate result( right );
	result -= left;
	result.Negate();

	return result;
}


inline Accuracy::Approximate const operator*( Accuracy::Approximate const& left, Accuracy::Approximate const& right ) {

	Accuracy::Approximate result( left );
	result *= right;

	return result;
}


inline Accuracy::Approximate const operator*( Accuracy::Approximate const& left, double const right ) {

	Accuracy::Approximate result( left );
	result *= right;

	return result;
}


inline Accuracy::Approximate const operator*( double const left, Accuracy::Approximate const& right ) {

	Accuracy::Approximate result( right );
	result *= left;

	return result;
}


inline Accuracy::Approximate const operator/( Accuracy::Approximate const& left, Accuracy::Approximate const& right ) {

	Accuracy::Approximate result( left );
	result /= right;

	return result;
}


inline Accuracy::Approximate const operator/( Accuracy::Approximate const& left, double const right ) {

	Accuracy::Approximate result( left );
	result /= right;

	return result;
}


inline Accuracy::Approximate const operator/( double const left, Accuracy::Approximate const& right ) {

	double value = std::numeric_limits< double >::quiet_NaN();
	int log2UpperBound = 0;
	int log2Epsilon    = 0;

	int const log2LowerBound = Accuracy::FloatTraits< double >::Exponent( right.GetValue() );
	if ( LIKELY( log2LowerBound > right.GetLog2Epsilon() ) ) {

		int const term = right.GetLog2Epsilon() + Accuracy::FloatTraits< double >::Exponent( left ) + 1;

		value = left / right.GetValue();

		log2UpperBound = Accuracy::FloatTraits< double >::Exponent( value ) + 1;

		log2Epsilon = 1 + term - 2 * log2LowerBound;
	}

	return Accuracy::Approximate( value, log2UpperBound, log2Epsilon );
}


inline std::ostream& operator<<( std::ostream& out, Accuracy::Approximate const& approximate ) {

	return( out << approximate.GetValue() << "±2^" << approximate.GetLog2Epsilon() );
}




//============================================================================
//    inline helpers
//============================================================================


inline Accuracy::Approximate const Square( Accuracy::Approximate const& value ) {

	Accuracy::Approximate result( value );
	result.Square();
	return result;
}


inline Accuracy::Approximate const Pow( Accuracy::Approximate const& value, int const power ) {

	if ( power < 0 )
		return( static_cast< double >( 1 ) / Pow( value, -power ) );

	Accuracy::Approximate result( 1 );
	if ( power > 0 ) {

		Accuracy::Approximate accumulator( value );
		unsigned int bits = power;
		for ( unsigned int bit = 1; bits != 0; bit += bit, accumulator *= accumulator ) {

			if ( ( bits & bit ) != 0 ) {

				result *= accumulator;
				bits &= ~bit;
			}
		}
	}
	return result;
}




//============================================================================
//    std inline helpers
//============================================================================


namespace std {


inline ::Accuracy::Approximate const abs( ::Accuracy::Approximate const& value ) {

	return ::Accuracy::Approximate( std::abs( value.GetValue() ), value.GetLog2UpperBound(), value.GetLog2Epsilon() );
}


inline ::Accuracy::Approximate const min( ::Accuracy::Approximate const& left, ::Accuracy::Approximate const& right ) {

	switch( ::Accuracy::Approximate::Compare( left, right ) ) {

		case -1: return left;
		case  1: return right;
		default: return ::Accuracy::Approximate( std::min( left.GetValue(), right.GetValue() ), std::max( left.GetLog2UpperBound(), right.GetLog2UpperBound() ), std::max( left.GetLog2Epsilon(), right.GetLog2Epsilon() ) );
	}
}


inline ::Accuracy::Approximate const max( ::Accuracy::Approximate const& left, ::Accuracy::Approximate const& right ) {

	switch( ::Accuracy::Approximate::Compare( left, right ) ) {

		case -1: return right;
		case  1: return left;
		default: return ::Accuracy::Approximate( std::max( left.GetValue(), right.GetValue() ), std::max( left.GetLog2UpperBound(), right.GetLog2UpperBound() ), std::max( left.GetLog2Epsilon(), right.GetLog2Epsilon() ) );
	}
}


}    // namespace std




#endif    /* __cplusplus */

#endif    /* __ACCURACY_APPROXIMATE_HH__ */
